# Based on
# https://api.semanticscholar.org/api-docs/graph#tag/Paper-Data/operation/get_graph_get_paper_citations

import json
from typing import Optional, List, Dict, Any
from urllib3.util.retry import Retry

import requests

OLD_API_URL_TEMPLATE = "https://api.semanticscholar.org/v1/paper/{paper_id}"
GRAPH_URL_TEMPLATE = "https://api.semanticscholar.org/graph/v1/paper/{paper_id}/citations?fields={fields}&offset={offset}&limit={limit}"
FIELDS = "title,authors,externalIds,venue,citationCount,publicationDate"


def _get_results(url: str) -> requests.Response:
    retry_strategy = Retry(
        total=3,
        backoff_factor=30,
        status_forcelist=[429, 500, 502, 503, 504],
        allowed_methods=["GET"],
    )

    session = requests.Session()
    adapter = requests.adapters.HTTPAdapter(max_retries=retry_strategy)
    session.mount("https://", adapter)

    try:
        response = session.get(url, timeout=30)
        response.raise_for_status()
        return response
    except (
        requests.exceptions.ConnectionError,
        requests.exceptions.RequestException,
        requests.exceptions.HTTPError,
    ) as e:
        print(f"Failed after {retry_strategy.total} retries: {str(e)}")
        raise

    return response


def _format_authors(authors: List[Dict[str, Any]]) -> List[str]:
    return [a["name"] for a in authors]


def _clean_entry(entry: Dict[str, Any]) -> Dict[str, Any]:
    entry = entry["citingPaper"]
    external_ids = entry.get("externalIds")
    if not external_ids:
        external_ids = dict()
    external_ids.pop("CorpusId", None)
    arxiv_id = external_ids.pop("ArXiv", None)
    return {
        "arxiv_id": arxiv_id,
        "external_ids": external_ids if external_ids else None,
        "title": entry["title"],
        "authors": _format_authors(entry["authors"]),
        "venue": entry.get("venue", ""),
        "citation_count": entry.get("citationCount", 0),
        "publication_date": entry.get("publicationDate", ""),
    }


def _format_entries(
    entries: List[Dict[str, Any]],
    start_index: int,
    total_results: int,
) -> str:
    clean_entries = [_clean_entry(e) for e in entries]
    return json.dumps(
        {
            "total_count": total_results,
            "returned_count": len(entries),
            "offset": start_index,
            "results": clean_entries,
        },
        ensure_ascii=False,
    )


def s2_citations(
    arxiv_id: str,
    offset: Optional[int] = 0,
    limit: Optional[int] = 50,
) -> str:
    """
    Get all papers that cited a given arXiv paper based on Semantic Scholar info.

    Returns a JSON object serialized to a string. The structure is:
    {"total_count": ..., "returned_count": ..., "offset": ..., "results": [...]}
    Every item in the "results" has the following fields:
    ("arxiv_id", "external_ids", "title", "authors", "venue", "citation_count", "publication_date")
    Use `json.loads` to deserialize the result if you want to get specific fields.

    Args:
        arxiv_id: The ID of a given arXiv paper.
        offset: The offset to scroll through citations. 10 items will be skipped if offset=10. 0 by default.
        limit: The maximum number of items to return. limit=50 by default.
    """

    assert isinstance(arxiv_id, str), "Error: Your arxiv_id must be a string"
    if "v" in arxiv_id:
        arxiv_id = arxiv_id.split("v")[0]
    paper_id = f"arxiv:{arxiv_id}"

    url = GRAPH_URL_TEMPLATE.format(paper_id=paper_id, fields=FIELDS, offset=offset, limit=limit)
    response = _get_results(url)
    result = response.json()
    entries = result["data"]
    total_count = len(result["data"]) + result["offset"]

    if "next" in result:
        paper_url = OLD_API_URL_TEMPLATE.format(paper_id=paper_id)
        paper_response = _get_results(paper_url)
        paper_result = paper_response.json()
        total_count = paper_result["numCitedBy"]

    return _format_entries(entries, offset if offset else 0, total_count)



import os
from typing import List, Any, Dict, cast, Optional, Callable
from dotenv import load_dotenv

from pydantic import BaseModel
from openai import OpenAI
from openai.types.chat import ChatCompletionMessageParam, ChatCompletionMessage


load_dotenv()

BASE_URL = os.getenv("BASE_URL", "https://openrouter.ai/api/v1")
API_KEY = os.getenv("OPENROUTER_API_KEY", "")
MODEL_NAME = os.getenv("DOCUMENT_QA_MODEL_NAME", "gpt-4o-mini")
SYSTEM_PROMPT = (
    "You are a helpful assistant that answers questions about documents accurately and concisely."
)
PROMPT = """Please answer the following questions based solely on the provided document.
If there is no answer in the document, output "There is no answer in the provided document".
First cite ALL relevant document fragments, then provide a final answer.
Answer all given questions one by one.
Make sure that you answer the actual questions, and not some other similar questions.

Questions:
{questions}

Document:
==== BEGIN DOCUMENT ====
{document}
==== END DOCUMENT ====

Questions (repeated):
{questions}

Your citations and answers:"""


class ChatMessage(BaseModel):  # type: ignore
    role: str
    content: str | List[Dict[str, Any]]


ChatMessages = List[ChatMessage]


def create_document_qa_func(
    base_url: str = BASE_URL,
    api_key: str = API_KEY,
    model_name: str = MODEL_NAME,
) -> Callable[..., Any]:
    def document_qa(
        questions: Optional[str] = None,
        document: Optional[str] = None,
        question: Optional[str] = None,
    ) -> str:
        """
        Answer questions about a document.
        Use this tool when you need to find relevant information in a big document.
        It takes questions and a document as inputs and generates an answer based on the document.

        Example:
        >>> document = "The quick brown fox jumps over the lazy dog."
        >>> answer = document_qa(questions="What animal is mentioned? How many of them?", document=document)
        >>> print(answer)
        "The document mentions two animals: a fox and a dog. 2 animals."

        Returns an answer to all questions based on the document content.

        Args:
        questions: Questions to be answered about the document.
        document: The full text of the document to analyze.
        question: Alias for 'questions'
        """

        if question and not questions:
            questions = question
        assert questions and questions.strip(), "Please provide non-empty 'questions'"
        assert document and document.strip(), "Please provide non-empty 'document'"

        messages: ChatMessages = [
            ChatMessage(role="system", content=SYSTEM_PROMPT),
            ChatMessage(
                role="user",
                content=PROMPT.format(questions=questions, document=document),
            ),
        ]

        sdk_messages = [
            cast(ChatCompletionMessageParam, m.model_dump(exclude_none=True)) for m in messages
        ]

        client = OpenAI(base_url=base_url, api_key=api_key)
        response: ChatCompletionMessage = (
            client.chat.completions.create(
                model=MODEL_NAME,
                messages=sdk_messages,
                temperature=0.0,
            )
            .choices[0]
            .message
        )

        if response.content is None:
            raise Exception("Response content is None")
        return response.content.strip()

    return document_qa



import json

from academia_mcp.tools import hf_datasets_search


def test_hf_datasets_search_gazeta() -> None:
    response = json.loads(hf_datasets_search(query="gazeta"))
    assert response["results"]
    assert "IlyaGusev/gazeta" in str(response)


def test_hf_datasets_search_cifar() -> None:
    response = json.loads(hf_datasets_search(query="CIFAR-10"))
    assert response["results"]
    assert "uoft-cs/cifar10" in str(response)



from academia_mcp.tools import arxiv_download


def test_arxiv_download() -> None:
    result = arxiv_download("2409.06820")
    assert "pingpong" in result.lower()


def test_arxiv_download_pdf() -> None:
    result = arxiv_download("2401.12474")
    assert "ditto" in result.lower()


def test_arxiv_download_bug() -> None:
    paper = arxiv_download("2409.14913v2")
    assert "Performance improves incrementally upon frontier model releases" in paper

    paper = arxiv_download("2501.08838v1")
    assert "ToMATO" in paper

    paper = arxiv_download("2501.06964v1")
    assert "Patient-Centric" in paper

    paper = arxiv_download("2412.08389v1")
    assert "enhance the efficacy of ESC systems" in paper



import os

import fire  # type: ignore
import uvicorn
from mcp.server.fastmcp import FastMCP
from dotenv import load_dotenv

from .tools.arxiv_search import arxiv_search
from .tools.arxiv_download import arxiv_download
from .tools.s2_citations import s2_citations
from .tools.hf_datasets_search import hf_datasets_search
from .tools.anthology_search import anthology_search
from .tools.document_qa import create_document_qa_func

load_dotenv()

API_KEY = os.getenv("OPENROUTER_API_KEY", "")
BASE_URL = os.getenv("BASE_URL", "https://openrouter.ai/api/v1")
MODEL_NAME = os.getenv("DOCUMENT_QA_MODEL_NAME", "gpt-4o-mini")


def run(
    host: str = "0.0.0.0",
    port: int = 5050,
    api_key: str = API_KEY,
    base_url: str = BASE_URL,
    model_name: str = MODEL_NAME,
) -> None:
    server = FastMCP("Academia MCP", stateless_http=True)

    server.add_tool(arxiv_search)
    server.add_tool(arxiv_download)
    server.add_tool(s2_citations)
    server.add_tool(hf_datasets_search)
    server.add_tool(anthology_search)

    if api_key:
        server.add_tool(
            create_document_qa_func(api_key=api_key, base_url=base_url, model_name=model_name)
        )

    http_app = server.streamable_http_app()
    uvicorn.run(http_app, host=host, port=port)


if __name__ == "__main__":
    fire.Fire(run)



# Based on
# https://github.com/jonatasgrosman/findpapers/blob/master/findpapers/searchers/arxiv_searcher.py
# https://info.arxiv.org/help/api/user-manual.html

import json
import re
from typing import Optional, List, Dict, Any, Union
from datetime import datetime, date
from urllib3.util.retry import Retry

import requests
import xmltodict

BASE_URL = "http://export.arxiv.org"
URL_TEMPLATE = "{base_url}/api/query?search_query={query}&start={start}&sortBy={sort_by}&sortOrder={sort_order}&max_results={limit}"
SORT_BY_OPTIONS = ("relevance", "lastUpdatedDate", "submittedDate")
SORT_ORDER_OPTIONS = ("ascending", "descending")


def _format_text_field(text: str) -> str:
    return " ".join([line.strip() for line in text.split() if line.strip()])


def _format_authors(authors: Union[List[Dict[str, str]], Dict[str, str]]) -> str:
    if not authors:
        return ""
    if isinstance(authors, dict):
        authors = [authors]
    names = [author["name"] for author in authors]
    result = ", ".join(names[:3])
    if len(names) > 3:
        result += f", and {len(names) - 3} more authors"
    return result


def _format_categories(categories: Union[List[Dict[str, Any]], Dict[str, Any]]) -> str:
    if not categories:
        return ""
    if isinstance(categories, dict):
        categories = [categories]
    clean_categories = [c.get("@term", "") for c in categories]
    clean_categories = [c.strip() for c in clean_categories if c.strip()]
    return ", ".join(clean_categories)


def _format_date(date: str) -> str:
    dt = datetime.strptime(date, "%Y-%m-%dT%H:%M:%SZ")
    return dt.strftime("%B %d, %Y")


def _clean_entry(entry: Dict[str, Any]) -> Dict[str, Any]:
    return {
        "id": entry["id"].split("/")[-1],
        "title": _format_text_field(entry["title"]),
        "authors": _format_authors(entry["author"]),
        "abstract": _format_text_field(entry["summary"]),
        "published": _format_date(entry["published"]),
        "updated": _format_date(entry["updated"]),
        "categories": _format_categories(entry.get("category", {})),
        "comment": _format_text_field(entry.get("arxiv:comment", {}).get("#text", "")),
    }


def _convert_to_yyyymmddtttt(date_str: str) -> str:
    try:
        date_obj = datetime.strptime(date_str, "%Y-%m-%d")
        return date_obj.strftime("%Y%m%d") + "0000"
    except ValueError as e:
        raise ValueError("Invalid date format. Please use YYYY-MM-DD format.") from e


def _has_cyrillic(text: str) -> bool:
    return bool(re.search("[а-яА-Я]", text))


def _compose_query(
    orig_query: str,
    start_date: Optional[str] = None,
    end_date: Optional[str] = None,
) -> str:
    query: str = orig_query.replace(" AND NOT ", " ANDNOT ")
    if "-" in query:
        query = f"({query}) OR ({query.replace('-', ' ')})"

    if start_date or end_date:
        if not start_date:
            start_date = "1900-01-01"
        if not end_date:
            today = date.today()
            end_date = today.strftime("%Y-%m-%d")
        date_filter = (
            f"[{_convert_to_yyyymmddtttt(start_date)} TO {_convert_to_yyyymmddtttt(end_date)}]"
        )
        query = f"({query}) AND submittedDate:{date_filter}"

    query = query.replace(" ", "+")
    query = query.replace('"', "%22")
    query = query.replace("(", "%28")
    query = query.replace(")", "%29")
    return query


def _format_entries(
    entries: List[Dict[str, Any]],
    start_index: int,
    include_abstracts: bool,
    total_results: int,
) -> str:
    clean_entries: List[Dict[str, Any]] = []
    for entry_num, entry in enumerate(entries):
        clean_entry = _clean_entry(entry)
        if not include_abstracts:
            clean_entry.pop("abstract")
        clean_entry["index"] = start_index + entry_num
        clean_entries.append(clean_entry)
    return json.dumps(
        {
            "total_count": total_results,
            "returned_count": len(entries),
            "offset": start_index,
            "results": clean_entries,
        },
        ensure_ascii=False,
    )


def _get_results(url: str) -> requests.Response:
    retry_strategy = Retry(
        total=3,
        backoff_factor=3,
        status_forcelist=[500, 502, 503, 504],
        allowed_methods=["GET"],
    )

    session = requests.Session()
    adapter = requests.adapters.HTTPAdapter(max_retries=retry_strategy)
    session.mount("http://", adapter)

    try:
        response = session.get(url, timeout=30)
        response.raise_for_status()
        return response
    except (
        requests.exceptions.ConnectionError,
        requests.exceptions.RequestException,
    ) as e:
        print(f"Failed after {retry_strategy.total} retries: {str(e)}")
        raise

    return response


def arxiv_search(
    query: str,
    offset: Optional[int] = 0,
    limit: Optional[int] = 5,
    start_date: Optional[str] = None,
    end_date: Optional[str] = None,
    sort_by: Optional[str] = "relevance",
    sort_order: Optional[str] = "descending",
    include_abstracts: Optional[bool] = False,
) -> str:
    """
    Search arXiv papers with field-specific queries.

    Fields:
        all: (all fields), ti: (title), au: (author),
        abs: (abstract), cat: (category), id: (ID without version)

    Operators:
        AND, OR, ANDNOT

    Please always specify the fields. Search should always be field-specific.
    You can search for an exact match of an entire phrase by enclosing the phrase in double quotes.
    If you do not need an exact match of a phrase, use single terms with OR/AND.
    Boolean operators are strict. OR is better in most cases.
    Do not include date constraints in the query: use "start_date" and "end_date" parameters instead.
    Use Latin script for names. For example, search "Ilya Gusev" instead of "Илья Гусев".

    Example queries:
        all:"machine learning"
        au:"del maestro"
        au:vaswani AND abs:"attention is all"
        all:role OR all:playing OR all:"language model"
        (au:vaswani OR au:"del maestro") ANDNOT ti:attention

    Returns a JSON object serialized to a string. The structure is:
    {"total_count": ..., "returned_count": ..., "offset": ..., "results": [...]}
    Every item in the "results" has the following fields:
    ("index", "id", "title", "authors", "abstract", "published", "updated", "categories", "comment")
    Use `json.loads` to deserialize the result if you want to get specific fields.

    Args:
        query: The search query, required.
        offset: The offset to scroll search results. 10 items will be skipped if offset=10. 0 by default.
        limit: The maximum number of items to return. limit=5 by default, limit=10 is the maximum.
        start_date: Start date in %Y-%m-%d format. None by default.
        end_date: End date in %Y-%m-%d format. None by default.
        sort_by: 3 options to sort by: relevance, lastUpdatedDate, submittedDate. relevance by default.
        sort_order: 2 sort orders: ascending, descending. descending by default.
        include_abstracts: include abstracts in the result or not. False by default.
    """

    assert isinstance(query, str), "Error: Your search query must be a string"
    assert isinstance(offset, int), "Error: offset should be an integer"
    assert isinstance(limit, int), "Error: limit should be an integer"
    assert isinstance(sort_by, str), "Error: sort_by should be a string"
    assert isinstance(sort_order, str), "Error: sort_order should be a string"
    assert query.strip(), "Error: Your query should not be empty"
    assert sort_by in SORT_BY_OPTIONS, f"Error: sort_by should be one of {SORT_BY_OPTIONS}"
    assert (
        sort_order in SORT_ORDER_OPTIONS
    ), f"Error: sort_order should be one of {SORT_ORDER_OPTIONS}"
    assert offset >= 0, "Error: offset must be 0 or positive number"
    assert limit < 100, "Error: limit is too large, it should be less than 100"
    assert limit > 0, "Error: limit should be greater than 0"
    assert not _has_cyrillic(query), "Error: use only Latin script for queries"
    assert include_abstracts is not None, "Error: include_abstracts must be bool"

    fixed_query: str = _compose_query(query, start_date, end_date)
    url = URL_TEMPLATE.format(
        base_url=BASE_URL,
        query=fixed_query,
        start=offset,
        limit=limit,
        sort_by=sort_by,
        sort_order=sort_order,
    )

    response = _get_results(url)
    content = response.content
    parsed_content = xmltodict.parse(content)

    feed = parsed_content.get("feed", {})
    total_results = int(feed.get("opensearch:totalResults", {}).get("#text", 0))
    start_index = int(feed.get("opensearch:startIndex", {}).get("#text", 0))
    entries = feed.get("entry", [])
    if isinstance(entries, dict):
        entries = [entries]
    formatted_entries: str = _format_entries(
        entries,
        start_index=start_index,
        total_results=total_results,
        include_abstracts=include_abstracts,
    )
    return formatted_entries



from .arxiv_search import arxiv_search
from .anthology_search import anthology_search
from .arxiv_download import arxiv_download
from .hf_datasets_search import hf_datasets_search
from .s2_citations import s2_citations
from .document_qa import create_document_qa_func


__all__ = [
    "arxiv_search",
    "arxiv_download",
    "anthology_search",
    "s2_citations",
    "hf_datasets_search",
    "create_document_qa_func",
]



import pytest

from academia_mcp.tools import anthology_search


def test_anthology_search_basic_search() -> None:
    result = anthology_search('ti:"BERT"')
    assert isinstance(result, str)
    assert len(result) > 0


def test_anthology_search_empty_query() -> None:
    with pytest.raises(AssertionError):
        anthology_search("")


def test_anthology_search_invalid_sort_by() -> None:
    with pytest.raises(AssertionError):
        anthology_search("abs:physics", sort_by="invalid_sort")


def test_anthology_search_invalid_sort_order() -> None:
    with pytest.raises(AssertionError):
        anthology_search("abs:physics", sort_order="invalid_order")


def test_anthology_search_negative_offset() -> None:
    with pytest.raises(AssertionError):
        anthology_search("abs:physics", offset=-1)


def test_anthology_search_negative_limit() -> None:
    with pytest.raises(AssertionError):
        anthology_search("abs:physics", limit=-1)


def test_anthology_search_zero_limit() -> None:
    with pytest.raises(AssertionError):
        anthology_search("abs:physics", limit=0)


def test_anthology_search_offset_limit_combination() -> None:
    result1 = anthology_search("abs:physics", offset=0, limit=5)
    result2 = anthology_search("abs:physics", offset=5, limit=5)
    assert result1 != result2


def test_anthology_search_sort_by_options() -> None:
    valid_sort_options = ["relevance", "published"]
    for sort_option in valid_sort_options:
        result = anthology_search("abs:physics", sort_by=sort_option)
        assert isinstance(result, str)
        assert len(result) > 0


def sort_order_options() -> None:
    result1 = anthology_search("abs:physics", sort_order="ascending")
    result2 = anthology_search("abs:physics", sort_order="descending")
    assert isinstance(result1, str)
    assert len(result1) > 0
    assert isinstance(result2, str)
    assert len(result2) > 0
    assert result1 != result2


def test_anthology_search_unicode_in_query() -> None:
    result = anthology_search('abs:"Schrödinger equation"')
    assert isinstance(result, str)
    assert len(result) > 0


def test_anthology_search_result_format() -> None:
    result = anthology_search("abs:physics")
    assert "title" in result.lower()


def test_anthology_search_type_validation() -> None:
    with pytest.raises(AssertionError):
        anthology_search(123)  # type: ignore
    with pytest.raises(AssertionError):
        anthology_search("abs:physics", offset="0")  # type: ignore
    with pytest.raises(AssertionError):
        anthology_search("abs:physics", limit="5")  # type: ignore
    with pytest.raises(AssertionError):
        anthology_search("abs:physics", sort_by=123)  # type: ignore
    with pytest.raises(AssertionError):
        anthology_search("abs:physics", sort_order=123)  # type: ignore


def test_anthology_search_integration_multiple_pages() -> None:
    first_page = anthology_search("abs:physics", offset=0, limit=5)
    second_page = anthology_search("abs:physics", offset=5, limit=5)
    third_page = anthology_search("abs:physics", offset=10, limit=5)
    assert first_page != second_page != third_page
    assert isinstance(first_page, str)
    assert isinstance(second_page, str)
    assert isinstance(third_page, str)


def test_anthology_search_start_date_only() -> None:
    result = anthology_search('au:wendler AND ti:"Do Llamas work"', start_date="2020-06-01")
    assert "English" in result


def test_anthology_find_conf() -> None:
    result = anthology_search('au:wendler AND ti:"Do Llamas work"')
    assert "2024.acl-long.820" in result



import json
from datetime import datetime
from typing import Optional, List, Dict, Any, Literal

from huggingface_hub import HfApi, DatasetInfo, hf_hub_download

HF_API = HfApi()


def _format_date(dt: Optional[datetime]) -> str:
    if not dt:
        return ""
    return dt.strftime("%B %d, %Y")


def _clean_entry(entry: DatasetInfo) -> Dict[str, Any]:
    try:
        readme_path = hf_hub_download(repo_id=entry.id, repo_type="dataset", filename="README.md")
        with open(readme_path, "r", encoding="utf-8") as f:
            readme_content = f.read()
    except Exception:
        readme_content = ""

    return {
        "id": entry.id,
        "created_at": _format_date(entry.created_at),
        "last_modified": _format_date(entry.last_modified),
        "downloads": entry.downloads,
        "likes": entry.likes,
        "tags": entry.tags,
        "readme": readme_content,
    }


def _format_entries(entries: List[DatasetInfo]) -> str:
    clean_entries: List[Dict[str, Any]] = [_clean_entry(entry) for entry in entries]
    return json.dumps({"results": clean_entries}, ensure_ascii=False)


def hf_datasets_search(
    query: Optional[str] = None,
    search_filter: Optional[List[str]] = None,
    limit: Optional[int] = 5,
    sort_by: Optional[str] = "trending_score",
    sort_order: Optional[str] = "descending",
) -> str:
    """
    Search or filter HF datasets.

    Examples:
        List only the datasets in Russian for language modeling:
        hf_datasets_search(filter=(language:ru", "task_ids:language-modeling"))

        List all recent datasets with "text" in their name
        hf_datasets_search(query="text", sort_by="last_modified")

    Returns a JSON object serialized to a string. The structure is: {"results": [...]}
    Every item in the "results" has the following fields:
    ("id", "created_at", "last_modified", "downloads", "likes", "tags")
    Use `json.loads` to deserialize the result if you want to get specific fields.

    Args:
        query: The search query for the exact match search.
        search_filter: A list of string to filter datasets.
        limit: The maximum number of items to return. limit=5 by default, limit=10 is the maximum.
        sort_by:
            The key with which to sort the resulting models.
            Possible values are "last_modified", "trending_score", "created_at", "downloads" and "likes".
            "trending_score" by default.
        sort_order: 2 sort orders: ascending, descending. descending by default.
    """
    direction: Optional[Literal[-1]] = -1 if sort_order == "descending" else None
    results = list(
        HF_API.list_datasets(
            search=query,
            sort=sort_by,
            direction=direction,
            limit=limit,
            filter=search_filter,
        )
    )
    return _format_entries(results)
